#ifndef _SCANDIR_HPP_
#define _SCANDIR_HPP_

#include <time.h>
#include "a_list.hpp"
#include "utils.hpp"
#include "fidoaddr.h"
#include "msg.hpp"
#include "msgbase.hpp"
#include "outbound.hpp"
#include "ufmtypes.h"
#include "parsetpl.hpp"
#include "tmstamp.hpp"

typedef enum { ACT_ERROR,
               ACT_MOVE, ACT_COPY, ACT_DELETE, ACT_REWRITE, ACT_NEWMSG,
               ACT_IGNORE, ACT_DISPLAY, ACT_ROUTE, ACT_FLAG, ACT_WRITEFILE,
               ACT_CALL, ACT_DELETEATTACH, ACT_CHANGEPATH, ACT_MOVEATTACH,
               ACT_SPLIT, ACT_ADDFILE, ACT_ADDNOTE, ACT_DELFILE, ACT_POLL,
               ACT_COPYATTACH
} tAct;

typedef enum { MASK_BODY, MASK_MASK, MASK_KLUDGE } MaskType;


class Mask {
public:
   
   MaskType _Type;
   FA _FromAddr;
   FA _ToAddr;
   char *_FromName;
   char *_ToName;
   char *_Subject;
   uint Loops;
   uint _Lines;
   uint _Bytes;
   ScanDir *sd;
   int  _SMask;
// ----- bit field attributes of message ----
   uint  fMaxAge      :2; // Max Age.
   uint  fMaxAttach   :2; // Max Attach Size
   uint  fEchomail    :2; // Message is echomail.
   uint  fScanned     :2; // Scanned
   uint  fLok         :2; // Locked
   uint  fAS          :2; // Archive/Sent
   uint  fLoop        :2; // Loop check
   uint  fPrivate     :2; // Private.
   uint  fCrash       :2; // Crash.
   uint  fReceived    :2; // Received
   uint  fSend        :2; // Send
   uint  fFileAttach  :2; // File attach
   uint  fTransit     :2; // Transit
   uint  fOrphan      :2; // Orphan
   uint  fKillSend    :2; // Kill/Send
   uint  fLocal       :2; // Local
   uint  fHold        :2; // Hold
   uint  fFileRequest :2; // File Request
   uint  fRRQ         :2; // RRQ
   uint  fIRR         :2; // IRR
   uint  fARQ         :2; // ARQ
   uint  fFURQ        :2; // FURQ
   uint  fDIR         :2; // DIR
   uint  fIMM         :2; // IMM
   uint  fCFM         :2; // CFM
   uint  fEmpty       :2; // Empty message (not have Body)
   uint  fTFS         :2; // TFS
   uint  fKFS         :2; // KFS
   uint  fAttExists   :2; // All attached files is exists

   Mask();
   ~Mask();
   void Print(void) const;
   int operator == ( cMSG &m) const;
};   

class Action {
public:
   FA   _f;
   tAct _Act;
   char *_TplName;
   PKTMode _Flav;
   Template *_Tpl;
   char *_OutDir;
   MSGBASE *_Base;
   Mask   *_Mask;
   int    _Lines;
   ScanDir *sd;
   ScanDir *Before;
   ScanDir *After;
   Action();
   ~Action();
   void Print(void);
   int Do(MSGBASE &b, cMSG &m);
};   

class DoList {
   IndBiList<Mask> Masks;
   IndBiList<Action> Actions;
public:
   DoList();
   ~DoList();
   void AddAction(Action &Act);
   void AddMask(Mask &Msk);
   void Print(void);
   int Do(MSGBASE &b, cMSG &m);
   Mask *InMask(cMSG &m);
   
};

class ScanDir {
public:
   IndBiList<DoList> _DoLists;
   IndBiList<tTimes> _Times;
   MSGBASE *_Base;
   int     _Renumber;
   int     _Unpack;
   uint    _MaxAge;
   uint    _MaxAttachSize;
   char    *_LoopStr;
   char    *_FlagFile;
   ScanDir();
   ~ScanDir();
   void SetBase(MSGBASE *B) { _Base = B; };
   MSGBASE *Base(void) { return _Base; };
   void Print(void);
   int Do(void);
   int DoWithRoute(MSGBASE &b, cMSG &m);
   int Execute(MSGBASE &b, cMSG &m);
   int Flagged(void);
inline   uint MaxAge(void) { if (_MaxAge == 0) { return ::MaxAge; } else { return _MaxAge;}; };
inline   uint MaxAttachSize(void) { if (_MaxAttachSize == 0) return ::MaxAttachSize; else return _MaxAttachSize; };
inline   char *LoopStr(void) { if (_LoopStr == NULL) return ::LoopStr; else return _LoopStr; };

};

extern IndBiList<ScanDir> ScanDirs;

int AddScanDir(void);
void PrintScanDirs(void);
void DoScanDirs(void);
int AddMask(void);
int AddBodyMask(void);
int AddKludgeMask(void);
int AddSMask(void);
int AddSBodyMask(void);
int AddSKludgeMask(void);
int AddAction(void);

#endif

