// --------------------------------------------------------------------
//       B a d M s g . c p p
//                                                                     
//        Fido messages tracker                            
//        Work with bad message.
// --------------------------------------------------------------------
//        Copyright (c) 1998,99 by Fyodor Ustinov                         
//                              FIDONet 2:5020/79                      
//                                                                     
//        All rights reserved.                                         
// --------------------------------------------------------------------
#ifndef UNIX
#include <io.h>
#else
#define stricmp strcasecmp
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "constant.hpp"
#include "help.hpp"
#include "utils.hpp"
#include "vars.hpp"
#include "cfg.hpp"
#include "scandir.hpp"
#include "ufmtypes.h"

// --------------------------------------------------------------------
// --------------------------------------------------------------------

int SetBadMode(void) {
static int BadAlready = FALSE;
char *tmt;

   if (BadAlready) {
      CfgError("BadMessages: already defined.");
      return(-1);
   }
   BadAlready = TRUE;
   tmt = GetOneWord(" ");
   if (strlen(tmt) == 0) {
      CfgError("Missed <Mode> part.");
      return (-1);
   }
   if (stricmp(tmt,"DELETE") == 0) {
      BadMsgMode = REMOVE;
   } else if (stricmp(tmt,"SKIP") == 0) {
      BadMsgMode = SKIP;
   } else if (stricmp(tmt,"EXIT") == 0) {
      BadMsgMode = EXIT;
   } else if (stricmp(tmt,"MOVE") == 0) {
      BadMsgMode = MOVE;
      tmt = GetStr();
      if (strlen(tmt) == 0) {
         CfgError("Missed Message base name.");
         return (-1);
      }
      if (*tmt == '$') {
         CfgError("You can't use Squish base as base for bad mail.");
         return (-1);
      }
    
      BadMsgBase = MakeBase(tmt);
      if (BadMsgBase == NULL) return (-1);
      if (!BadMsgBase->CheckOut()) {
         delete BadMsgBase;
         return (-1);
      }
      if (!BadMsgBase->Set(tmt,BASE_OUT)) {
         delete BadMsgBase;
         CfgError("Invalid message base name.");
         return (-1);
      }
      BadMsgBase->Rewind();
   } else {
      CfgError("Unknown <Mode>.");
      return (-1);
   }

  return (0);
}

void DoBadMsg(MSGBASE &b) {
char *Buff;
   switch(BadMsgMode) {
      case REMOVE:
         Log.Level(LOGE) << "Delete bad message " << b.MessageName() << EOL;
         b.DeleteMsg();
         break;
      case EXIT:
         Log.Level(LOGE) << "Message " << b.MessageName() << " is bad. Exit." << EOL;
         exit(-1);
      case SKIP:
         Log.Level(LOGE) << "Skip bad message " << b.MessageName() << EOL;
         break;
      case MOVE:
         Buff = b.ReadToMem();
         if (Buff == NULL) {
            Log.Level(LOGE) << "Skip bad message " << b.MessageName() << EOL;
            break;
         } else {
            Log.Level(LOGE) << "Move bad message " << b.MessageName() 
                            << " to " << BadMsgBase->BaseName() << EOL;
         }
         if (!BadMsgBase->WriteFromMem(Buff)) {
            free(Buff);
            break;
         }
         free(Buff);
         b.DeleteMsg();
         break;
   }
}

// ---------------------------- END --------------------------------------
