// --------------------------------------------------------------------
//       c f g . c p p
//                                                                     
//        Fido messages tracker                            
//        Work with configuration file
// --------------------------------------------------------------------
//        Copyright (c) 1998,99 by Fyodor Ustinov                         
//                              FIDONet 2:5020/79                      
//                                                                     
//        All rights reserved.                                         
// --------------------------------------------------------------------
#ifndef UNIX
#include <io.h>
#else
#define stricmp strcasecmp
#endif
#if defined (__DJGPP__) || defined (__EMX__)
#include <unistd.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "log.hpp"
#include "constant.hpp"
#include "help.hpp"
#include "utils.hpp"
#include "vars.hpp"
#include "cfg.hpp"
#include "scandir.hpp"
#include "badmsg.hpp"
#include "aka.hpp"
#include "age.hpp"
#include "attach.hpp"
#include "outbound.hpp"
#include "passwd.hpp"
#include "ufmtypes.h"

// --------------------------------------------------------------------
//                         Local function prototypes
// --------------------------------------------------------------------

int SetLogFile(void);
int SetLogLevel(void);
int SetNodelist(void);
int CommentLine(void);
int CommentCont(void);
int SetIndexFile(void);
int SetNodelistPath(void);
int SetMyAddr(void);
int SetSysopName(void);
int SetUseOwnZone(void);
int SetForceINTL(void);
int SetLoopStr(void);
int SetOrigin(void);
int SetTearline(void);
int SetNoLogIgnore(void);
int SetInclude(void);
int SetUTC(void);
int SetCreateMissingBase(void);
int SetAPktDir(void);
int SetMaxAttach(void);
int SetMaxNodelistAge(void);
int SetNewVia(void);
int SetSetViaAlways(void);
int SetSoftCheckInNodelists(void);
int SetCheckPoints(void);
// --------------------------------------------------------------------
//                                Local variables
// --------------------------------------------------------------------

static FILE *fh;

static CfgTok ConfigTokens[] = {
   {"LogFile:",&SetLogFile},
   {"LogLevel:",&SetLogLevel},
   {"Nodelist:",&SetNodelist},
   {"IndexFile:",&SetIndexFile},
   {"NodelistPath:",&SetNodelistPath},
   {"Address:",&SetMyAddr},
   {"SysopName:",&SetSysopName},
   {"UseOwnZone",&SetUseOwnZone},
   {"ForceINTL",&SetForceINTL},
   {"LoopStr:",&SetLoopStr},
   {"ScanDir:",&AddScanDir},
   {"Mask:",&AddMask},
   {"*Mask:",&AddSMask},
   {"BodyMask:",&AddBodyMask},
   {"*BodyMask:",&AddSBodyMask},
   {"KludgeMask:",&AddKludgeMask},
   {"*KludgeMask:",&AddSKludgeMask},
   {"Action:",&AddAction},
   {"MaxAge:",&SetMaxAge},
   {"BadMessages:",&SetBadMode},
   {"NoLogIgnore",&SetNoLogIgnore},
   {"Aka:",&SetAka},
   {"SkipHiddenFiles",&SetSkipHiddenFiles},
   {"FileInbound:",&SetFileInbound},
   {"Outbound:",&SetOutbound},
   {"AddToNullPkt",&SetAddToNullPkt},
   {"TempMail:",&SetTempMail},
   {"TrafficLog:",&SetTrafficLog},
   {"Password:",&SetPasswd},
   {"UTC:",&SetUTC},
   {"Origin:",&SetOrigin},
   {"Tearline:",&SetTearline},
   {"Include:",&SetInclude},
   {"APktDir:",&SetAPktDir},
   {"CreateMissingBase",&SetCreateMissingBase},
   {"MaxAttachSize:",&SetMaxAttach},
   {"KillRoutedMessage:",&SetKillRoutedMessages},
   {"MaxNodelistAge:",&SetMaxNodelistAge},
   {"UseNewVia",&SetNewVia},
   {"SetViaAlways",&SetSetViaAlways},
   {"SoftCheckInNodelists",&SetSoftCheckInNodelists},
   {"CheckPoints:",&SetCheckPoints},
   {"\\",&CommentLine},
   {"(",&CommentCont},
   {0,0}
};

// --------------------------------------------------------------------

int SetMaxAttach(void) {
int tmt;
IndBiList<ScanDir>::ElemPtr sd;

   sd = ScanDirs.GetLast();
   if (sd == NULL) {
      if (MaxAttachSize != 0) {
         CfgError("Global MaxAttachSize already defined.");
         return (-1);
      }
   } else {
      if (sd->_MaxAttachSize != 0) {
         CfgError("MaxAttachSize for this ScanDir already defined.");
         return (-1);
      }
   }
   tmt = GetLong();
   if (tmt < 1) {
      CfgError("Parameter must be a number great than 0.");
      return (-1);
   }
   if (sd != NULL) {
      sd->_MaxAttachSize = tmt;
   } else {
      MaxAttachSize = tmt;
   }
   return 0;
}

// --------------------------------------------------------------------

int SetNodelist(void) {
   return Ndl.AddNodelist();
}

// --------------------------------------------------------------------

int SetUseOwnZone(void) {
   if (UseOwnZone) {
      CfgError("UseOwnZone already set.");
      return (-1);
   }
   UseOwnZone = TRUE;
   return (0);
}

// --------------------------------------------------------------------

int SetNoLogIgnore(void) {
   if (!LogIgnore) {
      CfgError("NoLogIgnore already set.");
      return (-1);
   }
   LogIgnore = FALSE;
   return (0);
}

// --------------------------------------------------------------------

int SetCreateMissingBase(void) {
   if (CreateMissingBase) {
      CfgError("CreateMissingBase already set.");
      return (-1);
   }
   CreateMissingBase = TRUE;
   return (0);
}

// --------------------------------------------------------------------

int SetForceINTL(void) {
   if (ForceINTL) {
      CfgError("ForceINTL already set.");
      return (-1);
   }
   ForceINTL = TRUE;
   return (0);
}

// --------------------------------------------------------------------

int SetSoftCheckInNodelists(void) {
   if (SoftCheckInNodelists) {
      CfgError("SoftCheckInNodelists already set.");
      return (-1);
   }
   SoftCheckInNodelists = TRUE;
   return (0);
}

// --------------------------------------------------------------------

int SetNewVia(void) {
   if (NewVIAType) {
      CfgError("NewVIA already set.");
      return (-1);
   }
   NewVIAType = TRUE;
   return (0);
}

// --------------------------------------------------------------------

int SetSetViaAlways(void) {
   if (SetViaAlways) {
      CfgError("SetViaAlways already set.");
      return (-1);
   }
   SetViaAlways = TRUE;
   return (0);
}

// --------------------------------------------------------------------

int SetMyAddr(void) {
   if ((MyAddr.Zone() & FA_NOTDEF) != 0) {
      CfgError("Address already defined.");
      return (-1);
   }
   MyAddr = GetFA();
   if (!MyAddr.Valid()) {
      CfgError("Invalid FTN Address.");
      return (-1);
   }
   if (MyAddr.Masked()) {
      CfgError("You can't use mask in our address.");
      return (-1);
   }
   return (0);
}

// --------------------------------------------------------------------

int SetLoopStr(void) {
char *tmt;
IndBiList<ScanDir>::ElemPtr sd;

   sd = ScanDirs.GetLast();
   if (sd == NULL) {
      if (LoopStr != NULL) {
         CfgError("Loop string already defined.");
         return (-1);
      }
   } else {
      if (sd->_LoopStr != 0) {
         CfgError("Loop string for this ScanDir already defined.");
         return (-1);
      }
   }
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: Loop string.");
      return (-1);
   }
   if (sd != NULL) {
      sd->_LoopStr = strdup(tmt);
   } else {
      LoopStr = strdup(tmt);
   }
   return (0);
}

// --------------------------------------------------------------------

int SetAPktDir(void) {
char *tmt;
char Buff[1024];
char Buff2[1024];
   if (APktDir != NULL) {
      CfgError("APktDir already defined.");
      return (-1);
   }
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: directory name.");
      return (-1);
   }
   if (!DirExists(tmt)) {
      CfgError("Directory not exists.");
      return (-1);
   }

   if (Outbound != NULL) {
      strcpy(Buff,Outbound);
      if (Buff[strlen(Buff)-1] == PATHDELIMC) Buff[strlen(Buff)-1] = '\0';
      strcpy(Buff2,tmt);
      if (Buff2[strlen(Buff2)-1] == PATHDELIMC) Buff2[strlen(Buff2)-1] = '\0';
      if (stricmp(Buff,Buff2) == 0) {
         CfgError("APKT directory and Outbound directory should be in different place.");
         return (-1);
      }
   }

   APktDir = strdup(tmt);
   return (0);
}

// --------------------------------------------------------------------

int SetOrigin(void) {
char *tmt;
   if (Origin != NULL) {
      CfgError("Origin already defined.");
      return (-1);
   }
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: Origin string.");
      return (-1);
   }
   if (strlen(tmt) > 70) {
      CfgError("Origin string too long. Max Origin length is a 70 characters.");
      return (-1);
   }
   Origin = strdup(tmt);
   return (0);
}

// --------------------------------------------------------------------

int SetTearline(void) {
char *tmt;
   if (Tearline != NULL) {
      CfgError("Tearline already defined.");
      return (-1);
   }
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: Tearline string.");
      return (-1);
   }
   if (strlen(tmt) > 70) {
      CfgError("Tearline string too long. Max Tearline length is a 70 characters.");
      return (-1);
   }
   Tearline = strdup(tmt);
   return (0);
}

// --------------------------------------------------------------------

int CommentLine(void) {
   while (*CurrLinePos != '\0') CurrLinePos++;
   return (0);
}

// --------------------------------------------------------------------

static int CommentState = FALSE;

int CommentCont(void) {
   CommentState = TRUE;
   while (*CurrLinePos != '\0') CurrLinePos++;
   return (0);
}

// --------------------------------------------------------------------

int SetLogFile(void) {
char *tmt;

   if (LogFileName != NULL) {
      CfgError("Log file already defined.");
      return (-1);
   }
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: Log file name.");
      return (-1);
   }
   LogFileName = strdup(tmt);

   if (!GetLastTime(LogFileName)) {
      return (-1);
   }

   if (Log.Open(LogFileName) != 0) {
      return (-1);
   }

   Log.Level(LOGE) << "--- FTrack " << FVersion << " started." << EOL;

   return 0;
}

// --------------------------------------------------------------------

int SetSysopName(void) {
char *tmt;

   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: Sysop name.");
      return (-1);
   }
   SysopName = strdup(tmt);
   return 0;
}

// --------------------------------------------------------------------

int SetIndexFile(void) {
char *tmt;

   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: Index file name.");
      return (-1);
   }
   Ndl.IndexFile(tmt);
   return 0;
}

// --------------------------------------------------------------------

int SetNodelistPath(void) {
char *tmt;
char Buf[1024];

   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: Nodelists path.");
      return (-1);
   }
   if (!DirExists(tmt)) {
      CfgError("Directory not found.");
      return (-1);
   }
   strcpy(Buf,tmt);
   if (Buf[strlen(Buf)-1] != PATHDELIMC) {
      strcat(Buf,PATHDELIMS);
   }                                                                                                            
   NodelistPath = strdup(Buf);
   return 0;
}


// --------------------------------------------------------------------

int SetLogLevel(void) {
int tmt;

   if (LogLevel != -1) {
      CfgError("LogLevel already set.");
      return (-1);
   }
   tmt = GetInt();
   if (tmt < 0) {
      CfgError("Parameter must be a number between 0 and 65535.");
      return (-1);
   }
   Log.LogLevel(tmt);
   LogLevel = tmt;
   return 0;
}

// --------------------------------------------------------------------

int SetUTC(void) {
char *tmt;
char *tmt2;
   if (UTC != 99) {
      CfgError("UTC offset already defined.");
      return (-1);
   }
   tmt = GetOneWord(DELIMETERS);
   if (strlen(tmt) == 0) {
      CfgError("Missed UTC offset");
      return (-1);
   }
   tmt2 = tmt;
   if (*tmt2 != '+' && *tmt2 != '-' && isdigit((uchar)*tmt2) == 0) {
      CfgError("UTC offset must be a number between -12 and +12.");
      return (-1);
   }
   tmt2++;
   while (*tmt2 != '\0') {
      if (isdigit((uchar)*tmt2) == 0) {
         CfgError("UTC offset must be a number between -12 and +12.");
         return (-1);
      }
      tmt2++;
   }
   UTC = atoi(tmt);
   if (UTC < -12 || UTC > 12) {
      CfgError("UTC offset must be a number between -12 and +12.");
      return (-1);
   }
   return 0;
}

// --------------------------------------------------------------------

int SetCheckPoints(void){
char *tmt;

   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: <mode>.");
      return (-1);
   }
   if (stricmp(tmt,"NEVER") == 0) {
      CheckPoints = CHECKPNT_NEVER;
   } else if (stricmp(tmt,"HARD") == 0) {
      CheckPoints = CHECKPNT_HARD;
   } else if (stricmp(tmt,"SOFT") == 0) {
      CheckPoints = CHECKPNT_SOFT;
   } else {
      CfgError("Unknown CheckPoints mode.");
      return (-1);
   }

   return 0;
}


// --------------------------------------------------------------------
// --------------------------------------------------------------------

void AddEnv(char *Buf, char *env) {
char *tmt;
   tmt = getenv(env);
   if (tmt != NULL) {
      strcat(Buf,tmt);
   } else if (StrIsXNum(env)) {
      tmt = Buf+ strlen(Buf);
      *tmt++ = strtol(env,NULL,16);
      *tmt = '\0';
   } else {
//      strcat(Buf,env);
   }
}

// --------------------------------------------------------------------

void CfgError(char *msg) {
int i;
   Log.Level(LOGE) << "Error in '" << ConfigFile << "' Line: " << LastCfgLineNum << EOL;
   Log.Level(LOGE) << LastCfgLine << EOL;
   for (i = 0; i < (CurrLinePos - LastCfgLine); i++) {
      Log.Level(LOGE) << " ";
   }
   Log.Level(LOGE) << "^ ";
   Log.Level(LOGE) << msg << EOL;

}

// --------------------------------------------------------------------

char *GetStr(void){
// get string from config file.
// parse %f% as environment string
char *tmt, *tmt2, *c;
static char Buff[MAXCFGLINE];

   if (*CurrLinePos == '\"') {
      CurrLinePos++;
      tmt = GetOneWord("\"");
   } else {
      tmt = GetOneWord(DELIMETERS);
   }
   Buff[0] = '\0';
   tmt2 = Buff;
   while(*tmt != '\0') {
      if (*tmt == '%') {
         tmt++;
         if (*tmt == '%') {
            *tmt2++ = *tmt++;
            *tmt2 = '\0';
            continue;
         }

         if ((c = strchr(tmt,'%')) != NULL) {
            *c = '\0';
            AddEnv(tmt2,tmt);
            *c = '%';
            tmt = c + 1;
            tmt2 += strlen(tmt2);
         } else {
            *tmt2++ = '%';
            *tmt2++ = *tmt++;
            *tmt2 = '\0';
            continue;
         }
      } else {
         *tmt2++ = *tmt++;
         *tmt2 = '\0';
      }
   }
   return Buff;
}

// --------------------------------------------------------------------

FA &GetFA(void){
// get FidoAddress from config file.
char *tmt;
static FA fa;
   fa.Clean();
   tmt = GetStr();
//   tmt = GetOneWord(DELIMETERS);
   if (strlen(tmt) == 0) return fa;
   fa.Parse(tmt);
   return fa;
}

// --------------------------------------------------------------------

int GetInt(void){
// get next word as integer (0 - 65535)
char *tmt;
int result;
   tmt = GetOneWord(DELIMETERS);
   if (strlen(tmt) == 0) return(-3);
   if (!StrIsNum(tmt)) return (-1);
   result = atoi(tmt);
   if ((unsigned int) result > 65535U) {
      return (-2);
   } else {
      return result;
   }
}

// --------------------------------------------------------------------

int GetLong(void){
// get next word as long integer
char *tmt;
int result;
   tmt = GetOneWord(DELIMETERS);
   if (strlen(tmt) == 0) return(-3);
   if (!StrIsNum(tmt)) return (-1);
   result = atoi(tmt);
   return result;
}

// --------------------------------------------------------------------

int GetMInt(void){
// get next word as integer (0 - 65535), Asterisk - (-1) or bad (-2)
// ! - add 0x080000000
char *tmt;
int result;
   tmt = GetOneWord(DELIMETERS);
   if (strlen(tmt) == 0) return(-2);
   if (*tmt == '*') {
      return (-1);
   }
   result = 0;
   if (*tmt == '!') {
      result |= 0x08000000;
      tmt++;
   }
   if (!StrIsNum(tmt)) return (-2);
   result |= atoi(tmt);
   return result;
}

// --------------------------------------------------------------------

int GetOneLine (void) {
char *tmt;
   LastCfgLineNum++;
   if (fgets(LastCfgLine,MAXCFGLINE,fh) == NULL) {
      return (-1);
   }
// Strip end of line
   tmt = strrchr(LastCfgLine,'\n');
   if (tmt != NULL) *tmt = '\0';
   return 0;
}

// --------------------------------------------------------------------

char *GetOneWord(char *Delim) {
static char Buff[MAXCFGLINE];
int i;

   i = 0;
   Buff[0] = '\0';
   while (*CurrLinePos == ' ' || *CurrLinePos == '\t') CurrLinePos++;
   while (*CurrLinePos != '\0' && strchr(Delim,*CurrLinePos) == NULL) {
      Buff[i] = *CurrLinePos;
      CurrLinePos++;
      i++;
   }
   while (*CurrLinePos != '\0' && strchr(Delim,*CurrLinePos) != NULL) CurrLinePos++;
   while (*CurrLinePos == ' ' || *CurrLinePos == '\t') CurrLinePos++;
   Buff[i] = '\0';
   return Buff;
}

// --------------------------------------------------------------------

int ParseOneLine (void) {
char *tmt;
CfgTok *Tok;

   if (strlen(LastCfgLine) == 0) return(0);
   CurrLinePos = LastCfgLine;
   if (CommentState) {
      while (*CurrLinePos != '\0') {
         if (*CurrLinePos == ')') {
            CurrLinePos++;
            CommentState = FALSE;
            break;
         }
         CurrLinePos++;
      }
   }
   for (tmt = GetOneWord(" "); strlen(tmt) != 0; tmt = GetOneWord(" ")) {
      Tok = ConfigTokens;
      while (Tok->Token != 0) {
         if (stricmp(Tok->Token,tmt) == 0) {
            if (Tok->Func() != 0) {
               return (-1);
            } else {
               break;
            }
         }
         Tok++;
      }
      if (Tok->Token == 0) {
         CfgError("Unknown configuration word.");
         return (-1);
      }
   }
   return (0);
}


// --------------------------------------------------------------------

int ParseConfig(char *CfgFile) {

   CHP = 1003;
   if (access(CfgFile,R_OK) != 0) {
      fprintf(stderr,"Unable to open configuration file '%s'\n",CfgFile);
      return (-1);
   }
   CHP = 1004;
   fh = fopen(CfgFile,"rt");
   if (fh == 0) {
      fprintf(stderr,"Unable to open configuration file '%s'\n",CfgFile);
      return (-1);
   }

   CHP = 1005;
   LastCfgLine = (char *)malloc(MAXCFGLINE);
   CHP = 1006;
   CheckMem(LastCfgLine);
   CHP = 1007;
   LastCfgLineNum = 0;
   while (GetOneLine() == 0) {
      CHP = 1008;
      if (ParseOneLine() != 0) {
         CHP = 1009;
         free(LastCfgLine);
         fclose(fh);
         CHP = 100901;
         return (-1);
      }
      CHP = 1010;
   }
   if (!feof(fh)) {
      fprintf(stderr,"Error reading configuration file '%s'\n",CfgFile);
      return (-1);
   }
   free(LastCfgLine);
   LastCfgLine = NULL;
   fclose(fh);
   return (0);
}

// --------------------------------------------------------------------

int SetInclude(void) {
char *tmt;
int sLastCfgLineNum;
char *sLastCfgLine;
FILE *sFh;
char *sConfigFile;

   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: Include file name.");
      return (-1);
   }
   if (access(tmt,R_OK) != 0) {
      CfgError("File is not exists.");
      return (-1);
   }
   sLastCfgLine = LastCfgLine;
   sLastCfgLineNum = LastCfgLineNum;
   sFh = fh;
   sConfigFile = ConfigFile;
   ConfigFile = strdup(tmt);
   CHP = 1000;
   CheckMem(tmt);
   CHP = 1001;
   if (ParseConfig(tmt) != 0) {
      CHP = 1002;
      return(-1);
   }
   LastCfgLine = sLastCfgLine;
   LastCfgLineNum = sLastCfgLineNum;
   fh = sFh;
   free(ConfigFile);
   ConfigFile = sConfigFile;
   return (0);
}

// ---------------------------- END --------------------------------------
