/*--------------------------------------------------------------------*/
/*       L o g . c p p                                                */
/*                                                                    */
/*       Implementation of "Log" class                                */
/*--------------------------------------------------------------------*/
/*       Copyright (c) 1999 by Fyodor Ustinov                         */
/*                          FIDONet 2:5020/79                         */
/*                                                                    */
/*       All rights reserved.                                         */
/*--------------------------------------------------------------------*/

/*--------------------------------------------------------------------*/
/*                        System include files                        */
/*--------------------------------------------------------------------*/

#include <stdarg.h>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#ifndef UNIX
   #include <io.h>
#endif
#include <string.h>

/*--------------------------------------------------------------------*/
/*                         Local include files                        */
/*--------------------------------------------------------------------*/
#include "ufmtypes.h"
#include "log.hpp"

/*--------------------------------------------------------------------*/
/*                         Global definitions                         */
/*--------------------------------------------------------------------*/
// #define DEBUG_LOG 1


/*--------------------------------------------------------------------*/
/*                      Functions prototypes                          */
/*--------------------------------------------------------------------*/


/*--------------------------------------------------------------------*/
/*    T i m e                                                         */
/*                                                                    */
/*    Print current time                                              */
/*--------------------------------------------------------------------*/
char *CurrTime(void) {
time_t      td;
struct tm   *ts;
static char Buff[60];

    time(&td);
    ts = localtime(&td);
    strftime(Buff,60,"%X  ", ts);
    return (Buff);
}

/*--------------------------------------------------------------------*/
/*    D a t e                                                         */
/*                                                                    */
/*    Print current date                                              */
/*--------------------------------------------------------------------*/
char *Date(void) {
time_t      td;
struct tm   *ts;
static char Buff[60];

    time(&td);
    ts = localtime(&td);
    strftime(Buff,60,"%d.%m.%Y %H:%M:%S", ts);
    return (Buff);
}

/*--------------------------------------------------------------------*/
/*                       Class LogStream                              */
/*--------------------------------------------------------------------*/

LogStream::LogStream() {
   fh = NULL;
   NewLine = -1;
}

LogStream::~LogStream() {
   Close();
}

int LogStream::Open(char *FName) {
   if (strlen(FName) == 0) {
      fprintf(stderr,"Log file is not specified\n");
      return FALSE;
   }
   Close();
   fh = fopen(FName,"a+t");
   if (fh == NULL) {
      fprintf(stderr,"Unable to open log file '%s'\n",FName);
      return FALSE;
   }
   NewLine = -1;
   return (0);
}

int LogStream::Close(void) {
   if (fh != NULL) {
      int tmt;
      fflush(fh);
      tmt = fclose(fh);
      fh = NULL;
      return tmt;
   } else {
      return FALSE;
   }
}

void LogStream::LogLevel(uint Lvl) {
   _LogLevel = Lvl;
}

LogStream &LogStream::Level(uint Lvl) {
   cLogLevel = Lvl;
   return *this;
}

void LogStream::ShowLine(const char *format, ...) {
FILE *CurrFh;
bool ToFile, ToScreen;
va_list ll;

   va_start(ll, format);
   if (ll == NULL) return;
   ToFile = TRUE;
   ToScreen = FALSE;

   if (fh == NULL) {
      CurrFh = stdout;
   } else {
      CurrFh = fh;
   }
   if (cLogLevel < 100 && cLogLevel > _LogLevel) return;
   if (cLogLevel >= 200 && cLogLevel - 200 > _LogLevel) return;
   if (cLogLevel >= 100 && cLogLevel < 200 && cLogLevel - 100 > _LogLevel) return;
   if (cLogLevel >= 100) {
      ToScreen = TRUE;
   }
   if (cLogLevel >= 200) {
      ToFile = FALSE;
   }

   if (ToScreen) {
      vprintf(format,ll);
      fflush(stdout);
   }

   if (ToFile && CurrFh != stdout) {
      if (NewLine != 0)
         fprintf(CurrFh,"%s ",Date());
      vfprintf(CurrFh,format,ll);
      fflush(CurrFh);
   }
   if (NewLine != 0) NewLine = 0;

   va_end(ll);
}

LogStream &LogStream::operator << (const char *p) {

   ShowLine("%s", p);
   return *this;
}

LogStream &LogStream::operator << (char p) {

   ShowLine("%c", p);
   if (p == '\n') {
      NewLine = -1;
   }
   return *this;
}


LogStream &LogStream::operator << (signed int i) {

   ShowLine("%d", i);
   return *this;
}

LogStream &LogStream::operator<< (unsigned int i) {

   ShowLine("%u", i);
   return *this;
}

LogStream &LogStream::operator << (dword i) {

   ShowLine("%lu",i);
   return *this;
}

LogStream &LogStream::operator<< (sdword i) {

   ShowLine("%ld", i);
   return *this;
}



#ifdef DEBUG_LOG

LogStream aa;

void main(void) {
// LogStream aa("mylog.txt");

   aa.LogLevel(3);
   aa.Open("mylog.txt");
   aa.Level(1) << "Text: " << 1 << EOL;
   aa.Level(4) << 4;
   aa.Level(100) << 100;
   aa.Level(200) << 200;

}
#endif


/*      End of LOG.CPP      */


