// --------------------------------------------------------------------
//        S c a n d i r . c p p 
//                                                                     
//        Fido messages tracker                            
//        Scan dir class.
// --------------------------------------------------------------------
//        Copyright (c) 1998,99 by Fyodor Ustinov                         
//                              FIDONet 2:5020/79                      
//                                                                     
//        All rights reserved.                                         
// --------------------------------------------------------------------

#ifndef __GNUC__
   #if !defined(__QNX__) && !defined(__WATCOMC__)
      #include <io.h>
      #include <direct.h>
   #else
      #include <dirent.h>
   #endif
#else
   #include <unistd.h>
   #include <sys/types.h>
   #include <dirent.h>
   #include <sys/stat.h>
   #include <errno.h>
#endif                                                                                                          

#if defined(__QNX__) && defined(__WATCOMC__)
   #define strcasecmp stricmp
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <ctype.h>

#include "constant.hpp"
#include "vars.hpp"
#include "cfg.hpp"
#include "log.hpp"
#include "nodelist.hpp"
#include "utils.hpp"
#include "msg.hpp"
#include "msgbase.hpp"
#include "age.hpp"
#include "scandir.hpp"
#include "attach.hpp"
#include "aka.hpp"
#include "wildmat.hpp"
#include "msgapi.h"
#include "progprot.h"

// ---------------------------------------------------------
// ---------------------------------------------------------

Action::Action() {
   _Act = ACT_ERROR;
   _TplName = NULL;
   _OutDir = NULL;
   _Base = NULL;
   _Mask = NULL;
   _Tpl = NULL;
   sd = NULL;
   _f.Clean();
}
// --------------------------------------------------------------------

Action::~Action() {
   CHP = 1;
   if (_TplName != NULL) free(_TplName);
   if (_OutDir != NULL) free(_OutDir);
   if (_Base != NULL) delete _Base;
   if (_Mask != NULL) delete _Mask;
   if (_Tpl != NULL) delete _Tpl;
}

// --------------------------------------------------------------------

#include "scd_do.cpp"

// --------------------------------------------------------------------

void Action::Print(void) {
#ifdef _DEBUG

char *tmt;

   CHP = 22;
   switch(_Act) {
      case ACT_ERROR:
         tmt = "Error!";
         break;
      case ACT_IGNORE:
         tmt = "Ignore";
         break;
      default:
         tmt = "Unknown";
         break;
   }
   printf("Action: %s\n",tmt);
#endif
}

// ---------------------------------------------------------
// ---------------------------------------------------------

Mask::Mask() {
   CHP = 23;
   _FromName = NULL;
   _ToName = NULL;
   _Subject = NULL;
   Loops = (unsigned int)-1;
   sd = 0;
   _SMask = FALSE;
   fPrivate = fCrash = fReceived = fSend =
   fFileAttach = fTransit = fOrphan = fKillSend = fLocal =
   fHold = fFileRequest = fRRQ = fIRR = fARQ = fFURQ = fDIR =
   fIMM = fCFM = fTFS = fKFS  = fEmpty = fMaxAge = fLoop = fAttExists = 
   fEchomail = fScanned = fMaxAttach = fLok = fAS = 0;
}

Mask::~Mask() {
   CHP = 24;
   if (_FromName != NULL) free(_FromName);
   if (_ToName != NULL) free (_ToName);
   if (_Subject != NULL) free (_Subject);
}

void Mask::Print(void) const {
char fa1[40], fa2[40];

   if (_SMask) {
      Log << "*";
   }

   if (_Type == MASK_BODY) {
      Log << "BodyMask: \"" << _Subject << "\" " << _Lines << " " << _Bytes;
      return;
   }

   if (_Type == MASK_KLUDGE) {
      Log << "KludgeMask: \"" << _FromName << "\" \"" << _ToName << "\" " << _Lines;
      return;
   }
   CHP = 25;
   strcpy(fa1,_FromAddr.ToStr());
   strcpy(fa2,_ToAddr.ToStr());
   Log << "Mask: \"" << _FromName << "\" " << fa1
                   << " \"" << _ToName << "\" " << fa2 
                   << " \"" << _Subject << '"';
   fa1[0] = '\0';

   if (fMaxAge == 1) strcat(fa1,"+g");
   if (fMaxAge == 2) strcat(fa1,"-g");

   if (fAttExists == 1) strcat(fa1,"+x");
   if (fAttExists == 2) strcat(fa1,"-x");

   if (fPrivate == 1) strcat(fa1,"+p");
   if (fPrivate == 2) strcat(fa1,"-p");

   if (fCrash == 1) strcat(fa1,"+c");
   if (fCrash == 2) strcat(fa1,"-c");

   if (fReceived == 1) strcat(fa1,"+r");
   if (fReceived == 2) strcat(fa1,"-r");

   if (fSend == 1) strcat(fa1,"+s");
   if (fSend == 2) strcat(fa1,"-s");

   if (fFileAttach == 1) strcat(fa1,"+a");
   if (fFileAttach == 2) strcat(fa1,"-a");

   if (fTransit == 1) strcat(fa1,"+i");
   if (fTransit == 2) strcat(fa1,"-i");

   if (fOrphan == 1) strcat(fa1,"+o");
   if (fOrphan == 2) strcat(fa1,"-o");

   if (fKillSend == 1) strcat(fa1,"+k");
   if (fKillSend == 2) strcat(fa1,"-k");

   if (fLocal == 1) strcat(fa1,"+l");
   if (fLocal == 2) strcat(fa1,"-l");

   if (fHold == 1) strcat(fa1,"+h");
   if (fHold == 2) strcat(fa1,"-h");

   if (fFileRequest == 1) strcat(fa1,"+f");
   if (fFileRequest == 2) strcat(fa1,"-f");

   if (fRRQ == 1) strcat(fa1,"+q");
   if (fRRQ == 2) strcat(fa1,"-q");

   if (fIRR == 1) strcat(fa1,"+y");
   if (fIRR == 2) strcat(fa1,"-y");

   if (fARQ == 1) strcat(fa1,"+b");
   if (fARQ == 2) strcat(fa1,"-b");

   if (fFURQ == 1) strcat(fa1,"+u");
   if (fFURQ == 2) strcat(fa1,"-u");

   if (fDIR == 1) strcat(fa1,"+d");
   if (fDIR == 2) strcat(fa1,"-d");

   if (fIMM == 1) strcat(fa1,"+m");
   if (fIMM == 2) strcat(fa1,"-m");

   if (fEmpty == 1) strcat(fa1,"+e");
   if (fEmpty == 2) strcat(fa1,"-e");

   if (fTFS == 1) strcat(fa1,"+t");
   if (fTFS == 2) strcat(fa1,"-t");

   if (fKFS == 1) strcat(fa1,"+j");
   if (fKFS == 2) strcat(fa1,"-j");

   if (fCFM == 1) strcat(fa1,"+n");
   if (fCFM == 2) strcat(fa1,"-n");

   if (fEchomail == 1) strcat(fa1,"+E");
   if (fEchomail == 2) strcat(fa1,"-E");

   if (fMaxAttach == 1) strcat(fa1,"+A");
   if (fMaxAttach == 2) strcat(fa1,"-A");

   if (fScanned == 1) strcat(fa1,"+v");
   if (fScanned == 2) strcat(fa1,"-v");

   if (fLok == 1) strcat(fa1,"+L");
   if (fLok == 2) strcat(fa1,"-L");

   if (fAS == 1) strcat(fa1,"+S");
   if (fAS == 2) strcat(fa1,"-S");

   if (fLoop != 0) {
      if (fLoop == 1) strcat(fa1,"+");
      if (fLoop == 2) strcat(fa1,"-");
      fa2[0] = Loops + '0';
      fa2[1] = '\0';
      strcat(fa1,fa2);
   }
   if (fa1[0] == '\0') strcat(fa1,"*");
   Log << " " << fa1;
}

int SoftCmp(char *Str, char *Mask) {
// compare string with mask
int NotFlag;
int rc;
char *Str1;
char *Mask1, *m;
   CHP = 26;
   NotFlag = FALSE;
   if (strcmp(Mask,"*") == 0) return TRUE;
   if (Str == NULL) return FALSE;

   Str1 = strdup(Str);
   nls_strupr(Str1);
   Mask1 = strdup(Mask);
   nls_strupr(Mask1);
   m = Mask1;
   if (*Mask1 == '!') {
      NotFlag = TRUE;
      Mask1++;
   }
   if (*Mask1 == '~') {
      Mask1++;
      rc = (strstr(Str1,Mask1) != NULL);
// printf("-----1>Str == '%s' Mask == '%s' rc == %d\n",Str1,Mask1,rc);
      free(Str1);
      free(m);
      return (NotFlag != rc);
   } else {
      rc = (wildmat(Str1,Mask1) != 0);
// printf("-----2>Str == '%s' Mask == '%s' rc == %d\n",Str1,Mask1,rc);
      free(Str1);
      free(m);
      return (NotFlag != rc);
   }
}

int SoftKluCmp(cMSG &m, char *StrName, char *StrBody) {
// compare message kludges with mask
Kludge *Klu;
int rc;
int Comp;

   CHP = 26020;
   rc = 0;

   Klu = m._Klu.First();
   CHP = 26021;

   while (Klu != NULL) {
      CHP = 26022;
      Comp = FALSE;
      if (StrName == NULL) {
         CHP = 26023;
         if (Klu->Name() == NULL || strlen(Klu->Name()) == 0) Comp = TRUE;
      } else {
         CHP = 26024;
         if (SoftCmp(Klu->Name(),StrName)) Comp = TRUE;
      }
      CHP = 26025;
      if (Comp) {
         CHP = 26026;
         if (StrBody == NULL) {
            CHP = 26027;
            if (Klu->Body() == NULL || strlen(Klu->Body()) == 0) rc++;
         } else {
            CHP = 26028;
            if (SoftCmp(Klu->Body(),StrBody)) rc++;
         }
         CHP = 26029;
      }
      CHP = 26030;
      Klu = Klu->Next();
   }


   Klu = m._Via.First();
   CHP = 26031;

   while (Klu != NULL) {
      CHP = 26032;
      Comp = FALSE;
      if (StrName == NULL) {
         CHP = 26033;
         if (Klu->Name() == NULL || strlen(Klu->Name()) == 0) Comp = TRUE;
      } else {
         CHP = 26034;
         if (SoftCmp(Klu->Name(),StrName)) Comp = TRUE;
      }
      CHP = 26035;
      if (Comp) {
         CHP = 26036;
         if (StrBody == NULL) {
            CHP = 26037;
            if (Klu->Body() == NULL || strlen(Klu->Body()) == 0) rc++;
         } else {
            CHP = 26038;
            if (SoftCmp(Klu->Body(),StrBody)) rc++;
         }
         CHP = 26039;
      }
      CHP = 26040;
      Klu = Klu->Next();
   }
   CHP = 26041;
   return rc;
}

int EquFA(FA const &m, FA const &f) {
int NotFlag;
   CHP = 27;
   NotFlag = FALSE;
   if (m.Zone() & FA_NOTMASK) NotFlag = TRUE;
   if (m.Zone() & FA_OURMASK) {
      return (IsMyAka(f) != NotFlag);
   }
   if (m.Point() & FA_SUBMASK) {
      return (Ndl.InSubHubs(f,m) != NotFlag);
   }
   if (m.Zone() & FA_LSTMASK) {
      return ((Ndl.ExistInNodelist(f) != (unsigned int)-1) != NotFlag);
   }
   return(f == m);
}

int Mask::operator == (cMSG &m) const {
uint cMaxAge;
uint cLoops;
uint cMaxAttachSize;

   CHP = 28;
   if (LogLevel >= 5) {
      Log.Level(LOGD) << "Compare message with mask '";
      Print();
      Log.Level(LOGD) << "'" << EOL;
   }
   if (_Type == MASK_MASK) {
      if (!EquFA(_ToAddr, m._ToAddr)) return FALSE;
      if (!EquFA(_FromAddr, m._FromAddr)) return FALSE;
      if (!SoftCmp(m._FromName,_FromName)) return FALSE;
      if (!SoftCmp(m._ToName,_ToName)) return FALSE;
      if (!SoftCmp(m._Subject,_Subject)) return FALSE;
      if ((fPrivate - m.fPrivate) == 1) return FALSE;
      if ((fCrash - m.fCrash) == 1) return FALSE;
      if ((fReceived - m.fReceived) == 1) return FALSE;
      if ((fSend - m.fSend) == 1) return FALSE;
      if ((fFileAttach - m.fFileAttach) == 1) return FALSE;
      if ((fTransit - m.fTransit) == 1) return FALSE;
      if ((fOrphan - m.fOrphan) == 1) return FALSE;
      if ((fKillSend - m.fKillSend) == 1) return FALSE;
      if ((fLocal - m.fLocal) == 1) return FALSE;
      if ((fHold - m.fHold) == 1) return FALSE;
      if ((fFileRequest - m.fFileRequest) == 1) return FALSE;
      if ((fRRQ - m.fRRQ) == 1) return FALSE;
      if ((fIRR - m.fIRR) == 1) return FALSE;
      if ((fARQ - m.fARQ) == 1) return FALSE;
      if ((fFURQ - m.fFURQ) == 1) return FALSE;
      if ((fDIR - m.fDIR) == 1) return FALSE;
      if ((fIMM - m.fIMM) == 1) return FALSE;
      if ((fCFM - m.fCFM) == 1) return FALSE;
      if ((fTFS - m.fTFS) == 1) return FALSE;
      if ((fKFS - m.fKFS) == 1) return FALSE;
      if ((fEchomail - m.fEchomail) == 1) return FALSE;
      if ((fScanned - m.fScanned) == 1) return FALSE;
      if ((fLok - m.fLok) == 1) return FALSE;
      if ((fAS - m.fAS) == 1) return FALSE;
      if ((fEmpty - m.fEmpty) == 1) return FALSE;
      if (fLoop != 0) {
         cLoops = m.LoopCount(sd->LoopStr());
         if (fLoop == 1 && (cLoops - Loops > 0)) return FALSE;
         if (fLoop == 2 && (Loops - cLoops <= 0)) return FALSE;
      }
      cMaxAge = sd->MaxAge();
      if (fMaxAge == 1 && !AgeIsOver(m._Time,cMaxAge)) return FALSE;
      if (fMaxAge == 2 && AgeIsOver(m._Time,cMaxAge)) return FALSE;
      if (fAttExists != 0 || fMaxAttach != 0) {
         if (!m.fFileAttach) return FALSE;
         if (fAttExists == 1 && !AttachExists(m)) return FALSE;
         if (fAttExists == 2 && AttachExists(m)) return FALSE;
         cMaxAttachSize = m.AttachSize();
         if (fMaxAttach == 1 && (cMaxAttachSize <= sd->MaxAttachSize())) return FALSE;
         if (fMaxAttach == 2 && (cMaxAttachSize > sd->MaxAttachSize())) return FALSE;
      }
   } else if (_Type == MASK_BODY) {
      if (!SoftCmp(m.Body(),_Subject)) return FALSE;
      if (_Lines != (unsigned int)-1) {
         if (_Lines & 0x08000000) {
            if (m.Lines() >= (_Lines & 0x07ffffff)) return FALSE;
         } else {
            if (m.Lines() < (_Lines & 0x07ffffff)) return FALSE;
         }
      }
      if (_Bytes != (unsigned int)-1) {
         if (_Bytes & 0x08000000) {
            if (m.Bytes() >= (_Bytes & 0x07ffffff)) return FALSE;
         } else {
            if (m.Bytes() < (_Bytes & 0x07ffffff)) return FALSE;
         }
      }
   } else if (_Type == MASK_KLUDGE) {
      if ((cLoops = SoftKluCmp(m,_FromName,_ToName)) == 0) return FALSE;
      Log.Level(LOGD) << "Matches " << cLoops << " times." << EOL;
      if (_Lines != (unsigned int)-1) {
         if (_Lines & 0x08000000) {
            if (cLoops >= (_Lines & 0x07ffffff)) return FALSE;
         } else {
            if (cLoops < (_Lines & 0x07ffffff)) return FALSE;
         }
      }
   }
   CHP = 29;
   Log.Level(LOGD) << "Message is equal to mask." << EOL;
   return TRUE;
}


// ---------------------------------------------------------
// ---------------------------------------------------------

void DoList::AddAction(Action &a) {
   CHP = 30;
   Actions.AddToEnd(&a);
}

// ---------------------------------------------------------

void DoList::AddMask(Mask &m) {
   CHP = 31;
   Masks.AddToEnd(&m);
}

// ---------------------------------------------------------

DoList::DoList() {
   CHP = 32;
}

// ---------------------------------------------------------

DoList::~DoList() {
   CHP = 33;
   Actions.Clear();
   Masks.Clear();
}

// ---------------------------------------------------------

int DoList::Do(MSGBASE &b, cMSG &m) {
IndBiList<Action>::ElemPtr tmt;
 
   CHP = 34;
   for (tmt = Actions.GetFirst(); tmt != NULL; tmt++) {
      if (!tmt->Do(b,m)) {
         return FALSE;
      }
   }   
   return TRUE;
}

// ---------------------------------------------------------

Mask *DoList::InMask(cMSG &m) {
IndBiList<Mask>::ElemPtr tmt;

   CHP = 35;
   for (tmt = Masks.GetFirst(); tmt != NULL; tmt++) {
      if (*tmt == m) {
         return tmt;
      }
   }   
   return NULL;
}

// ---------------------------------------------------------

void DoList::Print(void) {
#ifdef _DEBUG
unsigned int i;
unsigned int Count;

   CHP = 36;
   Count = Masks.entries();
   for (i = 0; i < Count; i++) {
      Masks.find(i)->Print();
   }   

   Count = Actions.entries();
   for (i = 0; i < Count; i++) {
      Actions.find(i)->Print();
   }   
#endif
}

// ---------------------------------------------------------
// ---------------------------------------------------------

ScanDir::ScanDir() {
   CHP = 37;
   _Renumber = FALSE;
   _Unpack   = FALSE;
   _MaxAge   = 0;
   _MaxAttachSize = 0;
   _LoopStr = NULL;
   _FlagFile = NULL;
   _Base = NULL;
}

// ---------------------------------------------------------

ScanDir::~ScanDir() {
   CHP = 38;
   _DoLists.Clear();
   _Times.Clear();
   if (_LoopStr != NULL) {
      free(_LoopStr);
      _LoopStr = NULL;
   }
   if (_FlagFile != NULL) {
      free(_FlagFile);
      _FlagFile = NULL;
   }
   if (_Base != NULL) {
      delete _Base;
      _Base = NULL;
   }
}

// ---------------------------------------------------------

int ScanDir::Flagged(void) {
   if (_FlagFile == NULL) return TRUE;
   if (_FlagFile[0] == '!') {
      return (access(&_FlagFile[1],F_OK) != 0);
   } else {
      if (_FlagFile[0] == '#') {
         return (access(&_FlagFile[1],F_OK) == 0);
      } else {
         return (access(_FlagFile,F_OK) == 0);
      }
   }
}

// ---------------------------------------------------------

int ScanDir::Execute(MSGBASE &b, cMSG &m) {
IndBiList<DoList>::ElemPtr tmt;
Mask   *tmtm;

   CHP = 67;
   for (tmt = _DoLists.GetFirst(); tmt != NULL; tmt++) {
      CHP = 68;
      if ((tmtm = tmt->InMask(m)) != NULL) {
         CHP = 69;
         if (!tmt->Do(b,m)) {
            CHP = 70;
            return(FALSE);
         }
         CHP = 71;
         if (!tmtm->_SMask) {
            break;
         }
      }
      CHP = 72;
   }
   return TRUE;
}

// ---------------------------------------------------------

int ScanDir::Do(void) {
cMSG m;
MSGBASE *s;
   CHP = 39;
   if (_Base == NULL) return TRUE;
   if (_Base->Open() != TRUE) {
      return FALSE;
   }
   Log.Level(LOGD) << EOL;
   Log.Level(LOGI) << "Scanning message base " << _Base->BaseName() << EOL;
   Log.Level(LOGD) << "--------------------------------------" << EOL;
   if (!Flagged()) {
      Log.Level(LOGI) << "Wrong flag for this base." << EOL;
      _Base->Close();
      return TRUE;
   }
   if (!InTime(*this)) {
      Log.Level(LOGI) << "It's not time yet for this base." << EOL;
      _Base->Close();
      return TRUE;
   }
   if (_Unpack) {
      s = TempMail;
      Log.Level(LOGI) << "Unpack outbound to " << _Base->BaseName() << EOL;
      TempMail = _Base;
      if (!DoRepack()) {
         Log.Level(LOGE) << "Unable unpack outbound" << EOL;
         TempMail = s;
         _Base->Close();
         return FALSE;
      }
      TempMail = s;
   }

   if (!_Base->Rewind()) {
      Log.Level(LOGD) << "Base is empty." << EOL;
      _Base->Close();
      return TRUE;
   }
   CHP = 64;
   if (!_DoLists.IsEmpty()) {
      do {
         Log.Level(LOGD) << EOL;
         Log.Level(LOGD) << "Current message: " << _Base->MessageName() << EOL;
         if (!_Base->ReadMsg(m)) {
            if (errno == EACCES) {
               Log.Level(LOGE) << "Message " << _Base->MessageName() << " Locked by another process." << EOL;
               errno = 0;
            } else {
               CHP = 65;
               DoBadMsg(*_Base);
               CHP = 66;
            }
         } else {
            if (LogLevel >= 5) {
               m.Print();
            }
            if (Execute(*_Base,m) != TRUE) {
               _Base->Close();
               return FALSE;
            }
            CHP = 73;
         }
         CHP = 74;
      } while(_Base->Next());
      Log.Level(LOGD) << "No more messages." << EOL;
   } else {
      Log.Level(LOGD) << "No masks in scandir." << EOL;
   }
      
   if (_FlagFile != NULL && _FlagFile[0] == '#') {
      unlink(&_FlagFile[1]);
   }
   if (_Renumber) {
      Log.Level(LOGI) << "Renumber message base " << _Base->BaseName() << EOL;
      if (!_Base->Renumber()) {
         _Base->Close();
         return FALSE;
      }
   }
   _Base->Close();
   CHP = 75;
   return TRUE;
}

// ---------------------------------------------------------

int ScanDir::DoWithRoute(MSGBASE &b, cMSG &m) {
   CHP = 390001;
   if (_Base != NULL) {
      Log.Level(LOGE) << "Internal error. ScanDir::DoWithRoute _Base != NULL" << EOL;
      return FALSE;
   }
   Log.Level(LOGD) << EOL;
   Log.Level(LOGD) << "----- ScanDir::DoWithRoute -----" << EOL;
   if (!Flagged()) {
      Log.Level(LOGI) << "Wrong flag for this base." << EOL;
      return TRUE;
   }
   CHP = 390010;
   _Base = &b;
   if (!_DoLists.IsEmpty()) {
      Log.Level(LOGD) << EOL;
      if (LogLevel >= 5) {
         m.Print();
      }
      CHP = 390020;
      if (Execute(*_Base,m) != TRUE) {
         _Base = NULL;
         return FALSE;
      }
      CHP = 390030;
   } else {
      Log.Level(LOGD) << "No masks in scandir." << EOL;
   }
   _Base = NULL;
   if (_FlagFile != NULL && _FlagFile[0] == '#') {
      unlink(&_FlagFile[1]);
   }
   CHP = 390040;
   return TRUE;
}

// ---------------------------------------------------------

void ScanDir::Print(void) {
}

// ---------------------------------------------------------

IndBiList<ScanDir> ScanDirs;

void PrintScanDirs(void) {
IndBiList<ScanDir>::ElemPtr tmt;

   CHP = 41;
   for (tmt = ScanDirs.GetFirst(); tmt != NULL; tmt++) {
      tmt->Print();
   }   
}

void DoScanDirs(void) {
IndBiList<ScanDir>::ElemPtr tmt;

   CHP = 42;
   for (tmt = ScanDirs.GetFirst(); tmt != NULL; tmt++) {
      CHP = 60;
      if (!tmt->Do()) {
         return;
      }
      CHP = 61;
   }   
}

int AddScanDir(void) {
MSGBASE *Base;
ScanDir *sd;
char *tmt;
int Renumber, Unpack, br, ar;
char *FlagFile;

   CHP = 43;
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: Message base name.");
      return (-1);
   }
   br = ar = FALSE;
   if (stricmp(tmt,"@BeforeRoute") == 0) {
      Base = NULL;
      br = TRUE;
      
   } else if (stricmp(tmt,"@AfterRoute") == 0) {
      Base = NULL;
      ar = TRUE;
   } else {
      Base = MakeBase(tmt);
      if (Base == NULL) {
         return (-1);
      }
      if (!Base->Set(tmt,BASE_IN)) {
         delete Base;
         CfgError("Unable to open message base.");
         return (-1);
      }
      if (!Base->CheckIn()) {
         delete Base;
         return (-1);
      }
   }
   sd = new ScanDir();
   CheckMem((char*)sd);
   Renumber = FALSE;
   Unpack = FALSE;
   FlagFile = NULL;
   do {
      tmt = GetStr();
      if (strlen(tmt) != 0) {
         if (stricmp(tmt,"RENUMBER") == 0) {
            if (Renumber) {
               CfgError("Renumber for this base already set.");
               delete sd;
               delete Base;
               return (-1);
            }
            if (Base == NULL) {
               CfgError("Renumber for this base can not be set.");
               delete sd;
               return (-1);
            }
            Renumber = TRUE;
         } else if (stricmp(tmt,"Unpack") == 0) {
            if (Base == NULL) {
               CfgError("Unpack for this base can not be set.");
               delete sd;
               return (-1);
            }
            if (Unpack) {
               CfgError("Unpack for this base already set.");
               delete sd;
               delete Base;
               return (-1);
            }
            if (Outbound == NULL) {
               CfgError("Outbound directory not specified.");
               delete sd;
               delete Base;
               return (-1);
            }
            if (!Base->CheckOut()) {
               delete sd;
               delete Base;
               return (-1);
            }
            Unpack = TRUE;
         } else if (stricmp(tmt,"Flag") == 0) {
            tmt = GetStr();
            if (strlen(tmt) == 0) {
               CfgError("Missed flag file name.");
               delete sd;
               if (Base != NULL) delete Base;
               return (-1);
            }
            FlagFile = strdup(tmt);
         } else if (stricmp(tmt,"Daily") == 0) {
            if (Base == NULL) {
               CfgError("You can not use 'Daily' and 'Weekly' with this type of base.");
               delete sd;
               return (-1);
            }
            if (!ParseTime(*sd)) {
               delete sd;
               delete Base;
               return (-1);
            }
         } else if (stricmp(tmt,"Weekly") == 0) {
            if (Base == NULL) {
               CfgError("You can not use 'Daily' and 'Weekly' with this type of base.");
               delete sd;
               return (-1);
            }
            if (!ParseWeek(*sd)) {
               delete sd;
               delete Base;
               return (-1);
            }
         } else {
            CfgError("Unknown ScanDir parameter.");
            delete sd;
            if (Base != NULL) delete Base;
            return (-1);
         }
      }
   } while (strlen(tmt) != 0);


   sd->SetBase(Base);
   sd->_Renumber = Renumber;
   sd->_Unpack = Unpack;
   sd->_FlagFile = FlagFile;
   ScanDirs.AddToEnd(sd);
   if (br) BeforeRoute = sd;
   if (ar) AfterRoute = sd;
   return (0);
}

int ParseBodyMask(Mask &m) {
char *tmt;

   m._Type = MASK_BODY;
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: StringInBody.");
      return FALSE;
   }
//   if (strlen(tmt) > 72) {
//      CfgError("StringInBody too long. Max length of StringInBody is a 72 characters.");
//      return FALSE;
//   }
   m._Subject = strdup(tmt);
   m._Lines = GetMInt();
   if (m._Lines == (unsigned int)-2) {
      CfgError("Wrong Lines count. Should be '*' or positive integer.");
      return FALSE;
   }

   m._Bytes = GetMInt();
   if (m._Bytes == (unsigned int)-2) {
      CfgError("Wrong Bytes count. Should be '*' or positive integer.");
      return FALSE;
   }
   return TRUE;
}

int ParseKludgeMask(Mask &m) {
char *tmt;

   m._Type = MASK_KLUDGE;
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      m._FromName = NULL;
   } else {
      m._FromName = strdup(tmt);
   }
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      m._ToName = NULL;
   } else {
      m._ToName = strdup(tmt);
   }
   m._Lines = GetMInt();
   if (m._Lines == (unsigned int)-2) {
      CfgError("Wrong matches count. Should be '*' or positive integer.");
      return FALSE;
   }
   return TRUE;
}

int ParseMask(Mask &m, int Type) {
// Type == 0 - For mask.
// Type == 1 - For Rewrite;
char *tmt;
int FlagMode;
int OneFlag;
int rc;

   m._Type = MASK_MASK;
   CHP = 44;
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: 'From' name.");
      return FALSE;
   }
   m._FromName = strdup(tmt);

   m._FromAddr = GetFA();
   if (Type == 0) {
      rc = m._FromAddr.MaskValid(); 
   } else {
      rc = m._FromAddr.ActValid(); 
   }
   if (!rc) {
      CfgError("Invalid 'From' Address.");
      return FALSE;
   }
   if ((m._FromAddr.Zone() & FA_LSTMASK) && (!Ndl.Enabled())) {
      CfgError("You can't use '#' mask without nodelist.");
      return FALSE;
   }

   if ((m._FromAddr.Point() & FA_SUBMASK) && (!Ndl.Enabled())) {
      CfgError("You can't use '&' mask without nodelist.");
      return FALSE;
   }

   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed parameter: 'To' name.");
      return FALSE;
   }
   m._ToName = strdup(tmt);

   m._ToAddr = GetFA();
   if (Type == 0) {
      rc = m._ToAddr.MaskValid(); 
   } else {
      rc = m._ToAddr.ActValid(); 
   }
   if (!rc) {
      CfgError("Invalid 'To' Address.");
      return FALSE;
   }

   if ((m._ToAddr.Zone() & FA_LSTMASK) && (!Ndl.Enabled())) {
      CfgError("You can't use '#' mask without nodelist.");
      return FALSE;
   }

   if ((m._ToAddr.Point() & FA_SUBMASK) && (!Ndl.Enabled())) {
      CfgError("You can't use '&' mask without nodelist.");
      return FALSE;
   }

   tmt = GetStr();
//   if (strlen(tmt) == NULL) {
//      CfgError("Missed parameter: 'Subject'.");
//      return FALSE;
//   }
   m._Subject = strdup(tmt);

   FlagMode = 0;
   OneFlag = FALSE;
   while (*CurrLinePos != '\0') {
      if (isdigit((uchar)*CurrLinePos)) {
         if (Type != 0) {
            CfgError("You can use loop flag only in Mask.");
            return FALSE;
         }
         m.fLoop = (FlagMode == 2) ? 2 : 1;
         OneFlag = TRUE;
         if (m.Loops != (unsigned int)-1) {
             CfgError("Only one loop count should be in mask.");
             return FALSE;
         }
         m.Loops = *CurrLinePos - '0';
         CurrLinePos++;
         continue;
      }
      switch(*CurrLinePos) {
         case ' ':
         case '\t': // skip spaces and tabs
            break;
         case '-':
            FlagMode = 2;
            break;
         case '+':
            FlagMode = 1;
            break;
         case '*':
            CurrLinePos++;
            if (FlagMode != 0) {
               CfgError("You can't use '*' with '+' or '-'.");
               return FALSE;
            }
            return TRUE;
         case 'g': // Check Age
            OneFlag = TRUE;
            if (Type != 0) {
               CfgError("You can use 'g' flag only in Mask.");
               return FALSE;
            }
            if (m.sd->_MaxAge == 0 && MaxAge == 0) {
               CfgError("Global MaxAge is not defined.");
               return FALSE;
            }
            m.fMaxAge = (FlagMode == 2) ? 2 : 1;
            break;
         case 'A': // Check Max Attach size
            OneFlag = TRUE;
            if (Type != 0) {
               CfgError("You can use 'A' flag only in Mask.");
               return FALSE;
            }
            if (m.sd->_MaxAttachSize == 0 && MaxAttachSize == 0) {
               CfgError("Global MaxAttachSize is not defined.");
               return FALSE;
            }
            m.fMaxAttach = (FlagMode == 2) ? 2 : 1;
            break;
         case 'x': // Check existing attach
            OneFlag = TRUE;
            if (Type != 0) {
               CfgError("You can use 'x' flag only in Mask.");
               return FALSE;
            }
            m.fAttExists = (FlagMode == 2) ? 2 : 1;
            break;
         case 'E': // Check Echomail
            OneFlag = TRUE;
            if (Type != 0) {
               CfgError("You can use 'E' flag only in Mask.");
               return FALSE;
            }
            m.fEchomail = (FlagMode == 2) ? 2 : 1;
            break;
         case 'p': // Private
            OneFlag = TRUE;
            m.fPrivate = (FlagMode == 2) ? 2 : 1;
            break;
         case 'c': // Crash
            OneFlag = TRUE;
            m.fCrash = (FlagMode == 2) ? 2 : 1;
            break;
         case 'r': // Received
            OneFlag = TRUE;
            m.fReceived = (FlagMode == 2) ? 2 : 1;
            break;
         case 's': // Sent
            OneFlag = TRUE;
            m.fSend = (FlagMode == 2) ? 2 : 1;
            break;
         case 'a': // File Attach
            OneFlag = TRUE;
            m.fFileAttach = (FlagMode == 2) ? 2 : 1;
            break;
         case 'i': // Transit
            OneFlag = TRUE;
            m.fTransit = (FlagMode == 2) ? 2 : 1;
            break;
         case 'o': // Orphan
            OneFlag = TRUE;
            m.fOrphan = (FlagMode == 2) ? 2 : 1;
            break;
         case 'k': // Kill/Send
            OneFlag = TRUE;
            m.fKillSend = (FlagMode == 2) ? 2 : 1;
            break;
         case 'l': // Local
            OneFlag = TRUE;
            m.fLocal = (FlagMode == 2) ? 2 : 1;
            break;
         case 'h': // Hold
            OneFlag = TRUE;
            m.fHold = (FlagMode == 2) ? 2 : 1;
            break;
         case 'f': // File Request
            OneFlag = TRUE;
            m.fFileRequest = (FlagMode == 2) ? 2 : 1;
            break;
         case 'q': // RRQ
            OneFlag = TRUE;
            m.fRRQ = (FlagMode == 2) ? 2 : 1;
            break;
         case 'n': // CFM
            OneFlag = TRUE;
            m.fCFM = (FlagMode == 2) ? 2 : 1;
            break;
         case 'y': // IRR
            OneFlag = TRUE;
            m.fIRR = (FlagMode == 2) ? 2 : 1;
            break;
         case 'v': // Scanned
            OneFlag = TRUE;
            m.fScanned = (FlagMode == 2) ? 2 : 1;
            break;
         case 'L': // Lok
            OneFlag = TRUE;
            m.fLok = (FlagMode == 2) ? 2 : 1;
            break;
         case 'S': // A/S
            OneFlag = TRUE;
            m.fAS = (FlagMode == 2) ? 2 : 1;
            break;
         case 'b': // ARQ
            OneFlag = TRUE;
            m.fARQ = (FlagMode == 2) ? 2 : 1;
            break;
         case 'u': // Update Request
            OneFlag = TRUE;
            m.fFURQ = (FlagMode == 2) ? 2 : 1;
            break;
         case 'd': // Direct
            OneFlag = TRUE;
            m.fDIR = (FlagMode == 2) ? 2 : 1;
            break;
         case 'm': // IMM
            OneFlag = TRUE;
            m.fIMM = (FlagMode == 2) ? 2 : 1;
            break;
         case 'e': // Empty message
            if (Type != 0) {
               CfgError("You can use 'e' flag only in Mask.");
               return FALSE;
            }
            OneFlag = TRUE;
            m.fEmpty = (FlagMode == 2) ? 2 : 1;
            break;
         case 't': // TFS
            OneFlag = TRUE;
            m.fTFS = (FlagMode == 2) ? 2 : 1;
            break;
         case 'j': // KFS
            OneFlag = TRUE;
            m.fKFS = (FlagMode == 2) ? 2 : 1;
            break;
         default:
            if (!OneFlag) {
               CfgError("Missed or unknown 'Flag' mask.");
               return FALSE;
            } else {
               return TRUE;
            }
      }
      CurrLinePos++;
   }
   if (FlagMode == 0) {
      CfgError("Missed 'Flag' mask.");
      return FALSE;
   }
   return TRUE;
}   

static int PrevMask = 0;
DoList *LastDo;

int _AddMask(int SMask) {
Mask *m;
ScanDir *sd;

   CHP = 45;
   m = new Mask();
   sd = ScanDirs.GetLast();
   if (sd == NULL) {
      CfgError("Mask: without ScanDir:");
      delete m;
      return -1;
   }
   m->_SMask = SMask;
   m->sd = sd;
   if (!ParseMask(*m,0)) {
      delete m;
      return -1;
   }
   switch (PrevMask) {
      case 0: // first mask.
      case 1: // after action.
         LastDo = new DoList();
         sd->_DoLists.AddToEnd(LastDo);
         PrevMask = 2;
         break;
      case 2: // after MASK.
         break;
   }
   LastDo->AddMask(*m);
   return 0;
}

int AddMask(void) {
   return _AddMask(FALSE);
}

int AddSMask(void) {
   return _AddMask(TRUE);
}

int _AddBodyMask(int SMask) {
Mask *m;
ScanDir *sd;

   CHP = 45;
   m = new Mask();
   
   sd = ScanDirs.GetLast();
   if (sd == NULL) {
      CfgError("BodyMask: without ScanDir:");
      delete m;
      return -1;
   }
   m->sd = sd;
   m->_SMask = SMask;
   if (!ParseBodyMask(*m)) {
      delete m;
      return -1;
   }
   switch (PrevMask) {
      case 0: // first mask.
      case 1: // after action.
         LastDo = new DoList();
         sd->_DoLists.AddToEnd(LastDo);
         PrevMask = 2;
         break;
      case 2: // after MASK.
         break;
   }
   LastDo->AddMask(*m);
   return 0;
}

int AddBodyMask(void) {
   return _AddBodyMask(FALSE);
}

int AddSBodyMask(void) {
   return _AddBodyMask(TRUE);
}

int _AddKludgeMask(int SMask) {
Mask *m;
ScanDir *sd;

   CHP = 45001;
   m = new Mask();
   
   sd = ScanDirs.GetLast();
   if (sd == NULL) {
      CfgError("KludgeMask: without ScanDir:");
      delete m;
      return -1;
   }
   m->sd = sd;
   m->_SMask = SMask;
   if (!ParseKludgeMask(*m)) {
      delete m;
      return -1;
   }
   switch (PrevMask) {
      case 0: // first mask.
      case 1: // after action.
         LastDo = new DoList();
         sd->_DoLists.AddToEnd(LastDo);
         PrevMask = 2;
         break;
      case 2: // after MASK.
         break;
   }
   LastDo->AddMask(*m);
   return 0;
}

int AddKludgeMask(void) {
   return _AddKludgeMask(FALSE);
}

int AddSKludgeMask(void) {
   return _AddKludgeMask(TRUE);
}

// --------------------------------------------------------------------

int _AddIgnore(Action &a) {
   CHP = 46;
   a._Act = ACT_IGNORE;
   return 0;
}

int _AddDisplay(Action &a) {
char *tmt;
   CHP = 47;
   a._Act = ACT_DISPLAY;
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed String for displaying.");
      return(-1);
   }
   
   a._TplName = strdup(tmt);
   return 0;
}

int _AddWriteFile(Action &a, int i) {
char *tmt;
ScanDir *sd;

   CHP = 48;
   if (i == 1) {
      a._Act = ACT_WRITEFILE;
   } else {
      a._Act = ACT_ADDFILE;
   }
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed template name.");
      return(-1);
   }
   a._Tpl = new Template();
   if (!a._Tpl->Set(tmt)) {
      CfgError("Template file is not accesible.");
      delete a._Tpl;
      a._Tpl = NULL;
      return (-1);
   }
   sd = ScanDirs.GetLast();
   a._Tpl->sd = sd;

   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed file name.");
      delete a._Tpl;
      a._Tpl = NULL;
      return(-1);
   }
   a._OutDir = strdup(tmt);
   return 0;
}

int _AddCopy(Action &a) {
char *tmt;
   CHP = 49;
   a._Act = ACT_COPY;
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed message base name.");
      return(-1);
   }
   a._Base = MakeBase(tmt);
   if (a._Base == NULL) {
      return (-1);
   }
   if (!a._Base->Set(tmt,BASE_OUT)) {
      CfgError("Invalid message base name.");
      delete a._Base;
      a._Base = NULL;
      return(-1);
   }
   if (!a._Base->CheckOut()) {
      delete a._Base;
      a._Base = NULL;
      return (-1);
   }
   a._Base->Rewind();
   return 0;
}

int _AddMove(Action &a) {
char *tmt;
   CHP = 50;
   a._Act = ACT_MOVE;
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed message base name.");
      return(-1);
   }
   a._Base = MakeBase(tmt);
   if (a._Base == NULL) {
      return (-1);
   }
   if (!a._Base->Set(tmt,BASE_OUT)) {
      CfgError("Invalid message base name.");
      delete a._Base;
      a._Base = NULL;
      return(-1);
   }
   if (!a._Base->CheckOut()) {
      delete a._Base;
      a._Base = NULL;
      return (-1);
   }
   a._Base->Rewind();
   return 0;
}

int _AddDelete(Action &a) {
   CHP = 51;
   a._Act = ACT_DELETE;
   return 0;
}

int _AddRewrite(Action &a) {
Mask *m;

   CHP = 52;
   if (!a.sd->_Base->CheckOut()) {
      return -1;
   }
   a._Act = ACT_REWRITE;
   m = new Mask();
   if (!ParseMask(*m,1)) {
      delete m;
      return -1;
   }
   a._Mask = m;
   return 0;
}

int _AddNewMsg(Action &a) {
char *tmt;
Mask *m;
ScanDir *sd;
   CHP = 53;
   a._Act = ACT_NEWMSG;

   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed template name.");
      return(-1);
   }
   a._Tpl = new Template();
   if (!a._Tpl->Set(tmt)) {
      CfgError("Template file is not accesible.");
      delete a._Tpl;
      a._Tpl = NULL;
      return (-1);
   }
   sd = ScanDirs.GetLast();
   a._Tpl->sd = sd;

   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed message base name.");
      return(-1);
   }
   a._Base = MakeBase(tmt);
   if (a._Base == NULL) {
      return (-1);
   }
   if (!a._Base->Set(tmt,BASE_OUT)) {
      CfgError("Invalid message base name.");
      delete a._Tpl;
      a._Tpl = NULL;
      delete a._Base;
      a._Base = NULL;
      return(-1);
   }
   if (!a._Base->CheckOut()) {
      delete a._Tpl;
      a._Tpl = NULL;
      delete a._Base;
      a._Base = NULL;
      return (-1);
   }
   a._Base->Rewind();
   m = new Mask();
   if (!ParseMask(*m,1)) {
      delete m;
      delete a._Tpl;
      a._Tpl = NULL;
      delete a._Base;
      a._Base = NULL;
      return -1;
   }
   a._Mask = m;
   return 0;
}

int _AddDeleteAttach(Action &a) {
   CHP = 54;
   a._Act = ACT_DELETEATTACH;
   return 0;
}

int _AddChangePath(Action &a) {
char *tmt;
   CHP = 55;
   a._Act = ACT_CHANGEPATH;
   if (!a.sd->_Base->CheckOut()) {
      return -1;
   }
   tmt = GetStr();
   if (strlen(tmt) > 72) {
      CfgError("New path too long");
      return (-1);
   }
   a._OutDir = strdup(tmt);
   return 0;
}

int _AddMoveAttach(Action &a) {
char *tmt;
   CHP = 56;
   a._Act = ACT_MOVEATTACH;
   if (!a.sd->_Base->CheckOut()) {
      return -1;
   }
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed directory name.");
      return(-1);
   }
   if (!DirExists(tmt)) {
      CfgError("Target directory not found.");
      return(-1);
   }
   if (strlen(tmt) > 72) {
      CfgError("New path too long");
      return (-1);
   }
   a._OutDir = strdup(tmt);
   return 0;
}

int _AddCopyAttach(Action &a) {
char *tmt;
   CHP = 5601;
   a._Act = ACT_COPYATTACH;
   if (!a.sd->_Base->CheckOut()) {
      return -1;
   }
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed directory name.");
      return(-1);
   }
   if (!DirExists(tmt)) {
      CfgError("Target directory not found.");
      return(-1);
   }
   a._OutDir = strdup(tmt);
   return 0;
}

int _AddFlag(Action &a) {
char *tmt;
   CHP = 57;
   a._Act = ACT_FLAG;
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed file name.");
      return (-1);
   }
   a._OutDir = strdup(tmt);
   return 0;
}

int _AddDeleteFile(Action &a) {
char *tmt;
   CHP = 57;
   a._Act = ACT_DELFILE;
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed file name.");
      return (-1);
   }
   a._OutDir = strdup(tmt);
   return 0;
}


int _AddRoute(Action &a) {
char *tmt;

   CHP = 58;
   a._Act = ACT_ROUTE;
   if (a.sd == a.Before|| a.sd == a.After) {
      CfgError("You can not use the Action Route at usage ScanDir: @AfterRoute|@BeforeRoute");
      return (-1);
   }
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed Mode of route.");
      return (-1);
   }
   if (stricmp(tmt,"Direct") == 0) {
      a._Flav = DIRECT;
   } else if (stricmp(tmt,"Hold") == 0) {
      a._Flav = HOLD;
   } else if (stricmp(tmt,"Normal") == 0) {
      a._Flav = NORMAL;
   } else if (stricmp(tmt,"CRASH") == 0) {
      a._Flav = CRASH;
   } else {
      CfgError("Unknown Mode of route.");
      return (-1);
   }
   
   a._f = GetFA();
   if (a._f.Point() & FA_NOTDEF) {
      a._f.Point(0);
   }
   if (!a._f.RouValid()) {
      CfgError("Invalid routing address.");
      return (-1);
   }

   return 0;
}

int _AddPoll(Action &a) {
char *tmt;

   CHP = -58;
   a._Act = ACT_POLL;
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed Mode of poll.");
      return (-1);
   }
   if (stricmp(tmt,"Direct") == 0) {
      a._Flav = DIRECT;
   } else if (stricmp(tmt,"Hold") == 0) {
      a._Flav = HOLD;
   } else if (stricmp(tmt,"Normal") == 0) {
      a._Flav = NORMAL;
   } else if (stricmp(tmt,"CRASH") == 0) {
      a._Flav = CRASH;
   } else {
      CfgError("Unknown polling mode.");
      return (-1);
   }
   
   a._f = GetFA();
   if (a._f.Point() & FA_NOTDEF) {
      a._f.Point(0);
   }
   if (!a._f.RouValid()) {
      CfgError("Invalid polling address.");
      return (-1);
   }

   return 0;
}

int _AddCall(Action &a) {
char *tmt;
char *tmt2;
ScanDir *sd;
   a._Act = ACT_CALL;
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed programm name.");
      return(-1);
   }
   tmt = strdup(tmt);
   tmt2 = GetStr();
   if (strlen(tmt2) == 0) {
      a._TplName = tmt;
      return(0);
   }

   a._Tpl = new Template();
   if (!a._Tpl->Set(tmt)) {
      CfgError("Template file is not accesible.");
      delete a._Tpl;
      a._Tpl = NULL;
      return (-1);
   }
   sd = ScanDirs.GetLast();
   a._Tpl->sd = sd;

   a._OutDir = strdup(tmt2);
   free(tmt);

   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed programm name.");
      return(-1);
   }
   a._TplName = strdup(tmt);
   return 0;
}

int _AddSplit(Action &a) {
int i;

   a._Act = ACT_SPLIT;
   if (!a.sd->_Base->CheckOut()) {
      return -1;
   }
   i = GetInt();
   if (i < 1 || i > 65535) {
      CfgError("Parameter 'Lines' should be positive integer between 1 and 65535");
      return (-1);
   }
   a._Lines = i;
   return 0;
}

int _AddAddNote(Action &a) {
char *tmt;
ScanDir *sd;

   CHP = 53;
   a._Act = ACT_ADDNOTE;

   if (!a.sd->_Base->CheckOut()) {
      return -1;
   }
   tmt = GetStr();
   if (strlen(tmt) == 0) {
      CfgError("Missed template name.");
      return(-1);
   }
   a._Tpl = new Template();
   if (!a._Tpl->Set(tmt)) {
      CfgError("Template file is not accesible.");
      delete a._Tpl;
      a._Tpl = NULL;
      return (-1);
   }
   sd = ScanDirs.GetLast();
   a._Tpl->sd = sd;
   return 0;
}

// --------------------------------------------------------------------

int AddAction(void) {
Action *a;
char *tmt;
int rc;
ScanDir *sd;
   CHP = 59;
   
   if (PrevMask == 0) {
      CfgError("Action without Mask.");
      return(-1);
   }

   tmt = GetOneWord(" ");
   if (strlen(tmt) == 0) {
      CfgError("Missed action.");
      return (-1);
   }

   sd = ScanDirs.GetLast();
   PrevMask = 1;
   rc = 1;
   a = new Action();
   a->sd = sd;
   a->Before = BeforeRoute;
   a->After = AfterRoute;
   if (stricmp(tmt,"Ignore") == 0) {
      rc = _AddIgnore(*a);
   } else if (stricmp(tmt,"Delete") == 0) {
      rc = _AddDelete(*a);
   } else if (stricmp(tmt,"Display") == 0) {
      rc = _AddDisplay(*a);
   } else if (stricmp(tmt,"Move") == 0) {
      rc = _AddMove(*a);
   } else if (stricmp(tmt,"Copy") == 0) {
      rc = _AddCopy(*a);
   } else if (stricmp(tmt,"Rewrite") == 0) {
      rc = _AddRewrite(*a);
   } else if (stricmp(tmt,"WriteFile") == 0) {
      rc = _AddWriteFile(*a,1);
   } else if (stricmp(tmt,"AppendToFile") == 0) {
      rc = _AddWriteFile(*a,2);
   } else if (stricmp(tmt,"NewMsg") == 0) {
      rc = _AddNewMsg(*a);
   } else if (stricmp(tmt,"DeleteAttach") == 0) {
      rc = _AddDeleteAttach(*a);
   } else if (stricmp(tmt,"ChangePath") == 0) {
      rc = _AddChangePath(*a);
   } else if (stricmp(tmt,"MoveAttach") == 0) {
      rc = _AddMoveAttach(*a);
   } else if (stricmp(tmt,"CopyAttach") == 0) {
      rc = _AddCopyAttach(*a);
   } else if (stricmp(tmt,"Flag") == 0) {
      rc = _AddFlag(*a);
   } else if (stricmp(tmt,"Route") == 0) {
      rc = _AddRoute(*a);
   } else if (stricmp(tmt,"Call") == 0) {
      rc = _AddCall(*a);
   } else if (stricmp(tmt,"AddNote") == 0) {
      rc = _AddAddNote(*a);
   } else if (stricmp(tmt,"DelFile") == 0) {
      rc = _AddDeleteFile(*a);
   } else if (stricmp(tmt,"Poll") == 0) {
      rc = _AddPoll(*a);
   } else if (stricmp(tmt,"Split") == 0) {
      rc = _AddSplit(*a);
   }


   if (rc == 1) {
      CfgError("Unknown Action.");
      delete a;
      return (-1);
   }
   if (rc == -1) {
      delete a;
      return (-1);
   }
      
   LastDo->AddAction(*a);
   return 0;
}

#if 0
DoList dl;
void main(void) {
Action *a1;
Mask *m1;
   a1 = new Action(ACT_MOVE);
   m1 = new Mask();
   dl.AddMask(*m1);
   dl.AddAction(*a1);
   dl.Print();
}
#endif
