// --------------------------------------------------------------------
//        t m s t a m p . c p p
//                                                                     
//        Fido messages tracker                            
//        Check time of execution ScanDir's
// --------------------------------------------------------------------
//        Copyright (c) 1998,99 by Fyodor Ustinov                         
//                              FIDONet 2:5020/79                      
//                                                                     
//        All rights reserved.                                         
// --------------------------------------------------------------------
#ifndef __GNUC__
   #if !defined(__QNX__) && !defined(__WATCOMC__)
      #include <io.h>
      #include <direct.h>
   #else
      #include <dirent.h>
   #endif
#else
   #include <unistd.h>
   #include <sys/types.h>
   #include <dirent.h>
   #include <sys/stat.h>
   #include <errno.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "constant.hpp"
#include "help.hpp"
#include "utils.hpp"
#include "vars.hpp"
#include "cfg.hpp"
#include "scandir.hpp"
#include "age.hpp"
#include "ufmtypes.h"
#include "tmstamp.hpp"
#include "wildmat.hpp"

// --------------------------------------------------------------------
time_t LastTime = 0;
// --------------------------------------------------------------------
// --------------------------------------------------------------------

static time_t TimeOfBeginOfDay(int Day) {
time_t t;
struct tm *tmbuf;

   t = time(NULL);
   tmbuf = localtime(&t);
   tmbuf->tm_sec = 0;
   tmbuf->tm_min = 0;
   tmbuf->tm_hour = 0;
   t = mktime(tmbuf);
   if (Day == -1) {
      return t;
   }
   Day -= tmbuf->tm_wday;
//   if (Day < 0) {
//      Day += 7;
//   }
   t += Day * (24*60*60);
   return t;
}

// --------------------------------------------------------------------

int GetLastTime(char *FName) {
struct stat _fstat;

   if (access(FName,F_OK) != 0) {
      LastTime = 0;
      return TRUE;
   }
   if (stat(FName,&_fstat) != 0) {
      Log.Level(LOGE) << "Unable to get time from file '" << FName << '"' << EOL;
      return FALSE;
   }
   LastTime = _fstat.st_mtime;
   return TRUE;
}

// --------------------------------------------------------------------

int InTime(ScanDir &Dir) {
time_t CTime;
time_t ETime, STime;

IndBiList<tTimes>::ElemPtr tmt;
 
   tmt = Dir._Times.GetFirst(); 
   if (tmt == NULL) {
      return TRUE;
   }
   CTime = time(NULL);
   for (;tmt != NULL; tmt++) {                                                             
      ETime = tmt->_ETime;
      STime = tmt->_STime;
      Log.Level(LOGD) << "CTime == " << (int) CTime << " STime == " << (int) STime 
                      << " ETime == " << (int) ETime << " LastTime == " << (int) LastTime << EOL;
      if (ETime != 0) {
         if (CTime >= STime && CTime <= ETime) return TRUE;
      } else {
         if (LastTime <= STime && CTime >= STime) return TRUE;
      }
   }
   return FALSE;
}

// --------------------------------------------------------------------

static time_t GetTime(char *tmt) {
time_t t;   
int i;
   i = atoi(tmt);
   if (i < 0 || i > 23) {
      return (time_t)-1;
   }
   t = i * 60;
   tmt += 3;
   i = atoi(tmt);
   if (i < 0 || i > 59) {
      return (time_t)-1;
   }
   t += i;
   t *= 60;
   return t;
}

// --------------------------------------------------------------------
// --------------------------------------------------------------------

int ParseTime(ScanDir &Dir) {
// parse time format
// xx:xx
// xx:xx-
// xx:xx-xx:xx
char *tmt;
tTimes *t;

   if (LogFileName == NULL) {
      CfgError("You can't set scanning time without Log File.");
      return FALSE;
   }
   tmt = GetStr();
   do {
      t = new tTimes;
      if (strlen(tmt) == 0) {
         t->_STime = TimeOfBeginOfDay(-1);
         t->_ETime = 0;
         Dir._Times.AddToEnd(t);
         return TRUE;
      }
      if (wildmat(tmt,"##:##") != TRUE && wildmat(tmt,"##:##-") != TRUE &&
          wildmat(tmt,"##:##-##:##") != TRUE) {
         CfgError("Invalid time format");
         delete t;
         return FALSE;
      }
      t->_STime = GetTime(tmt);
      if (t->_STime == (time_t)-1) {
         CfgError("Invalid time format");
         delete t;
         return FALSE;
      }
      t->_STime += TimeOfBeginOfDay(-1);
      t->_ETime = 0;
      if (tmt[5] == '-') {
         if (tmt[6] != '\0') {
            tmt+=6;
            t->_ETime = GetTime(tmt);
            if (t->_ETime == (time_t)-1) {
               CfgError("Invalid time format");
               delete t;
               return FALSE;
            }
         } else {
            t->_ETime = (23*60*60)+(59*60)+59;
         }
         t->_ETime += TimeOfBeginOfDay(-1);
      }
      if (t->_ETime != 0 && t->_ETime < t->_STime) {
         time_t tmt;
         tmt = t->_ETime;
         t->_ETime = (23*60*60)+(59*60)+59;
         t->_ETime += TimeOfBeginOfDay(-1);
         Dir._Times.AddToEnd(t);
         t = new tTimes;
         t->_STime = TimeOfBeginOfDay(-1);
         t->_ETime = tmt;
      }
      Dir._Times.AddToEnd(t);
      tmt = GetStr();
   } while (strlen(tmt) != 0);
   return TRUE;
}

// --------------------------------------------------------------------

int ParseWeek(ScanDir &Dir) {
// parse week format
// x
// x-
// x-x
char *tmt;
tTimes *t;
int i;

   if (LogFileName == NULL) {
      CfgError("You can't set scanning time without Log File.");
      return FALSE;
   }
   tmt = GetStr();
   do {
      t = new tTimes;
      if (strlen(tmt) == 0) {
         t->_STime = TimeOfBeginOfDay(0);
         t->_ETime = 0;
         Dir._Times.AddToEnd(t);
         return TRUE;
      }
      if (wildmat(tmt,"#") != TRUE && wildmat(tmt,"#-") != TRUE &&
          wildmat(tmt,"#-#") != TRUE) {
         CfgError("Invalid day of week format");
         delete t;
         return FALSE;
      }
      i = atoi(tmt);
      if (i > 6) {
         CfgError("Invalid day of week format");
         delete t;
         return FALSE;
      }
      t->_STime = TimeOfBeginOfDay(i);
      t->_ETime = 0;
      if (tmt[1] == '-') {
         if (tmt[2] != '\0') {
            tmt+=2;
            i = atoi(tmt);
            if (i > 6) {
               CfgError("Invalid day of week format");
               delete t;
               return FALSE;
            }
            t->_ETime = TimeOfBeginOfDay(i);
         } else {
            t->_ETime = TimeOfBeginOfDay(6) + (23*60*60)+(59*60)+59;
         }
      }
      Dir._Times.AddToEnd(t);
      tmt = GetStr();
   } while (strlen(tmt) != 0);
   return TRUE;
}
// ---------------------------- END --------------------------------------
